import re
from typing import Tuple, List, Type, Set, Callable, Union, Any, Iterable, Optional, Dict
from logging import getLogger, Logger
import socket
import camel_case_switcher.dict_processor
import copy

from tornado.httputil import HTTPMessageDelegate, HTTPServerRequest
from tornado.routing import ReversibleRouter, ReversibleRuleRouter, PathMatches, Rule
from tornado.web import RequestHandler, Application

from . import IResponder, IRouter, IApplication
from . import HandlerType, HandlerListType, HandlerController
from . import BasicResponder
from . import Logged_RequestHandler, Empty_RequestHandler, HealthCheck_RequestHandler
from .tools import ExtendedLogger, RequestLogger
from .tools import ConfigLoader
from .tools import RegExpType

from tornado.httpserver import HTTPServer, HTTPRequest
from tornado.ioloop import IOLoop

class _Protocols:
    __http = 'http'
    __https = 'https'
    
    http = __http
    HTTP = __http
    
    https = __https
    HTTPS = __https

class ApplicationBase(IApplication):
    """
    ApplicationBase class.
    
    The following attributes can be overrided:
    :attribute logger_name
    :attribute handlers
    List of handlers those are passed to the to tornado.web.Application class during the initialisation.
    If untouched (None or missing), default handlers will be used.
    """
    
    default_handlers: HandlerListType = \
    [
        (r"^/?$", Empty_RequestHandler, dict(redirect_page='/static/index.html')),
        (r"^/healthcheck(|(/.*))$", HealthCheck_RequestHandler),
    ]
    custom_router: Union[IRouter, Type[IRouter], None] = None
    handler_controllers: List[HandlerController]
    responder_class: Type[IResponder] = BasicResponder
    hosts: List[str] = None
    logger_name: str = 'http_server'
    access_logger_name:  str = None
    bind_address: str = None
    new_format_access_log:  bool = False
    
    __access_logger: ExtendedLogger = None
    __server: HTTPServer = None
    __router: Optional[IRouter] = None
    __settings: Dict[str, Any] = None
    
    __self_addr: str = None
    __listen_port: int = None
    __protocol: str = None
    __host_params: Dict[str, Any] = None
    
    #region Initialization
    def __init__\
    (
        self,
        *,
        name: str = None,
        config_name: str = 'main',
        config_prefix: str = 'HTTP',
        config_priority: bool = False,
        hosts: List[str] = None,
        bind_address: str = None,
        handler_controllers: List[HandlerController] = None,
        handlers: HandlerListType = None,
        custom_router: Union[IRouter, Type[IRouter], None] = None,
        **settings,
    ):
        
        """
        Initializes a new instance of tornado.web.Application and prepares tornado.httpserver.HTTPServer to be run.
        :param name:
        str. Name of server used in serveral log records.
        
        :param config_name:
        Name of config in the ConfigLoader class, where the settings are loaded from.
        By default, settings are loaded from the main config.
        :param config_prefix:
        Prefix part of the config path to the server's settings. Path keys are '/'-separated.
        Note that due to the implementation's restrictions, server settings could not be in the top-level of the config.
        :param config_priority:
        By default, arguments passed directly to the initializer, have more priority.
        By setting config_config_priority=True, you are prioritising config over the keyword arguments.
        
        :param hosts
        List of str.
        :param bind_address
        str or None. The address the core server is bound.
        :param handler_controllers
        List of HandlerController initialized objects
        :param custom_router:
        Type of IRouter, or instantiated IRouter.
        A router that should be used in application.
        
        :param settings:
        Keyword arguments. Partially parsed by ApplicationBase, partially passed to the tornado.web.Application
        All ApplicationBase optional parameters are listed below.
        Note that all CamelCase parameters loaded from the config would be morphed into the underscore_style,
            so `selfAddress` and `ListenPort` are completely legal.
        
        :param static_files:
        Same as `static_path`.
        If both are presented, `static_path` has priority.
        
        :param self_address:
        str. Hardcoded server uri. Used only for the info message about server start and several responses based on the `self_address` property.
        Should contain protocol.
        :param listen_ip:
        str. Used only if `self_address` is not set up for the `self_address` calculation.
        If missing as well, server will try to self-discover.
        The resulting self_address will have protocol - http or https, ip-address and, if custom, a port.
        :param ip:
        Same as `listen_ip`.
        If both are presented, `listen_ip` has priority.
        :param listen_port:
        int value of port, which server is uses to listen requests.
        By default 80 or 443 port is used - according to the SSL configuration
        :param port:
        Same as `listen_port`.
        If both are presented, `listen_port` has priority.
        
        :param SSL
        NOT IMPLEMENTED YET
        
        dict. Group of parameters required for SSL.
        Any of them can be passed directly as an argument as well.
        """
        
        # TODO: SSL
        # TODO: Hosts description
        # TODO: Handlers description
        # TODO: Handler Controllers description
        
        if (name is not None):
            self.name = name
        if (not hasattr(self, 'name') or self.name is None):
            self.name = type(self).__name__
        if (handler_controllers is not None):
            self.handler_controllers = handler_controllers
        if (not hasattr(self, 'handler_controllers') or self.handler_controllers is None):
            self.handler_controllers = list()
        if (not hasattr(self, 'custom_router') or self.custom_router is None):
            self.custom_router = custom_router
        
        if (bind_address is not None):
            self.bind_address = bind_address
        if (not hosts is None):
            self.hosts = hosts
        if (not hasattr(self, 'hosts') or self.hosts is None):
            self.hosts = [ 'self' ]
        
        self.logger: ExtendedLogger = getLogger(self.logger_name)
        if (not hasattr(self, 'access_logger_name') or self.access_logger_name is None):
            self.access_logger_name = f'{self.logger_name}.access'
        _access_logger: ExtendedLogger = getLogger(self.access_logger_name)
        self.__access_logger: RequestLogger = RequestLogger(None, _access_logger)
        
        if (handlers is not None):
            self.handlers = handlers
            self.logger.debug(f"{self.name}: List of handlers: {handlers}")
        elif (getattr(self, 'handlers', None) is not None):
            self.logger.debug(f"{self.name}: List of handlers: {self.handlers}")
        else:
            self.handlers = ApplicationBase.default_handlers
            self.logger.debug(f"{self.name}: Using default handlers")
        
        _settings = camel_case_switcher.dict_processor.dict_keys_camel_case_to_underscope(ConfigLoader.get_from_config(config_prefix, config_name=config_name, default=lambda: dict()))
        if (config_priority):
            settings_deep_copy = copy.deepcopy(settings)
            settings_deep_copy.update(_settings)
            _settings = settings_deep_copy
        else:
            _settings.update(settings)
        
        if (_settings.get('static_files') and not _settings.get('static_path')):
            _settings['static_path'] = _settings['static_files']
        
        self.__host_params = self.initialize_host(_settings)
        if (self.custom_router is None):
            _h = self.handlers
        else:
            if (isinstance(self.custom_router, IRouter)):
                self.__router = self.custom_router
            elif (isinstance(self.custom_router, type) and issubclass(self.custom_router, IRouter)):
                self.__router = self.custom_router(self, settings.get('optimize_tree', False))
            else:
                raise ValueError(f"Invalid value of custom router: '{self.custom_router}' (type: {type(self.custom_router).__name__})")
            _h = [ self.__router ]
        
        self.__settings = _settings
        super(IApplication, self).__init__(handlers=_h, **_settings)
        self.attach_controllers(**self.__host_params)
    
    def initialize_host(self, settings):
        _ssl = None
        # _ssl = _settings.get('SSL')
        if (_ssl):
            protocol = _Protocols.HTTPS
        else:
            protocol = _Protocols.HTTP
        
        listen_port = settings.get('listen_port') or settings.get('port')
        if (listen_port is None):
            if (protocol == _Protocols.HTTP):
                listen_port = 80
            elif (protocol == _Protocols.HTTPS):
                listen_port = 443
            listen_port_part = ""
        else:
            listen_port_part = f":{listen_port}"
        
        self.__listen_port = listen_port
        for host in self.hosts:
            host_address = settings.get(f'{host}_address')
            if (host_address):
                listen_address = None
            else:
                if (hasattr(self, f'get_{host}_host') and callable(getattr(self, f'get_{host}_host'))):
                    _host = getattr(self, f'get_{host}_host')()
                    if (isinstance(_host, tuple) and len(_host) == 2):
                        listen_address, host_pattern = _host
                    else:
                        listen_address = _host
                else:
                    listen_address = \
                        settings.get('listen_address') \
                        or settings.get('listen_ip') \
                        or settings.get('ip') \
                        or socket.gethostbyname(socket.gethostname()) \
                
                host_address = f"{protocol}://{listen_address}{listen_port_part}"
            
            setattr(self, f'_{type(self).__name__}__{host}_addr', host_address)
            if (not hasattr(self, f'{host}_address')):
                setattr(type(self), f'{host}_address', host_address)
        
        return dict(protocol=protocol, listen_port_part=listen_port_part)
    
    def initialize_settings(self):
        pass
    #endregion
    
    #region Discovery
    @classmethod
    def __self_discover_generator(cls):
        names = \
        [
            'localhost',
            socket.getfqdn(),
        ]
        for name in names:
            try:
                for _value in cls.__self_discover_generator_per_interface(name=name):
                    yield _value
            except socket.herror:
                continue
        
        local_addr = cls.get_public_ip()
        try:
            for _value in cls.__self_discover_generator_per_interface(addr=local_addr):
                yield _value
        except socket.herror:
            pass
    
    @classmethod
    def __self_discover_generator_per_interface(cls, name=None, addr=None):
        if (addr is None):
            assert not name is None, "Either name or addr is required"
            addr = socket.gethostbyname(name)
        yield addr
        _name, _alias_list, _address_list = socket.gethostbyaddr(addr)
        yield _name
        for _alias in _alias_list:
            yield _alias
        for addr in _address_list:
            yield addr
            yield socket.getfqdn(_name)
            yield socket.gethostbyname(_name)
    
    @classmethod
    def get_public_ip(cls):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            s.connect(("8.8.8.8", 80))
            local_addr = s.getsockname()[0]
            s.close()
            return local_addr
        except OSError:
            return socket.gethostbyname(socket.gethostname())
    
    @classmethod
    def self_discover(cls) -> Set[str]:
        """
        Server will try to do some self-discovery job.
        Not used in the distributed version.
        
        :return:
        The result is non-sorted set of unique items which could be associated with the local host.
        Usually it looks like:
        { '127.0.0.1', '127.0.1.1', 'localhost', 'MyComputer', '192.168.0.14' }
        """
        
        return set(cls.__self_discover_generator())
    
    @property
    def listen_port(self) -> int:
        return self.__listen_port
    
    def get_self_host(self) -> Tuple[str, str]:
        return self.get_public_ip(), '.*'
    
    @property
    def self_address(self) -> str:
        """
        Returns the self_address value, which was calculated during the initialisation process.
        :return:
        Returns self_address value.
        """
        
        return getattr(self, f"_{type(self).__name__}__self_addr")
    #endregion
    
    #region Start server
    def run(self, blocking=True, num_processes=1):
        """
        Runs the server on the port specified earlier.
        Server blocks the IO.
        """
        
        self.logger.info(f"{self.name}: Starting HTTP service...")
        
        self.__server = HTTPServer(self)
        self.__server.listen(self.__listen_port, address=self.bind_address)
        self.logger.info(f"{self.name}: Service started")
        for host in self.hosts:
            host_address = getattr(self, f'{host}_address')
            self.logger.info(f"{self.name}: Listening on the {host} host: {host_address}")
        
        self.__server.start(num_processes=num_processes)
        
        if (blocking):
            IOLoop.current().start()
    
    @classmethod
    def simple_start_prepare(cls):
        from logging import getLogger
        from http_server_base.tools import setup_logging
        
        # Configure logger
        setup_logging()
        logger = getLogger(cls.logger_name)
        logger.setLevel('TRACE')
        logger.info("Logger started")
        ConfigLoader.load_configs(soft_mode=True)
        
        return logger
    
    @classmethod
    def __simple_start_func(cls, pre_run_func: Union[Callable[['ApplicationBase'], None], None], num_processes=1, blocking=True, **settings) -> Tuple[Union['ApplicationBase', None], Union[Exception, None]]:
        """
        If any expected error occurs, returns the instance of initialized class (if any) and the exception
        
        :param pre_run_func:
        Function to be run after the server initialization but before actual start.
        Must take only one argument - the server instance.
        :param settings:
        Settings used for the server initialization
        :return:
        Tuple:
         - instance of initialized object (if any)
         - exception info
        """
        
        server_application: ApplicationBase = cls(**settings)
        if (callable(pre_run_func)):
            pre_run_func(server_application)
        
        try:
            server_application.run(num_processes=num_processes, blocking=blocking)
        except KeyboardInterrupt:
            server_application.logger.info("Keyboard interrupt. Exiting now.")
        except PermissionError as e:
            return server_application, e
        
        return None, None
    
    @classmethod
    def simple_start_server(cls, pre_run_func: Union[Callable[['ApplicationBase'], None], None] = None, **settings):
        """
        Initializes logging and ConfigLoader in the soft-mode,
        than initializes the class instance, runs this function and starts the server.
        
        :param pre_run_func:
        Function to be run after the server initialization but before actual start.
        Must take only one argument - the server instance.
        :param settings:
        Settings used for the server initialization
        :return:
        """
        
        logger = cls.simple_start_prepare()
        
        try:
            server_application, exception = cls.__simple_start_func(pre_run_func, **settings)
            
            if (isinstance(exception, PermissionError)):
                server_application.logger.error("Permission error. Restarting server with the port of range 8xxx.")
                new_port = server_application.listen_port + 8000
                del server_application
                settings['listen_port'] = new_port
                logger.info(f"Restarting server on the port {new_port}.")
                cls.__simple_start_func(pre_run_func, **settings)
                
                
        except Exception:
            logger.exception(msg="Unhandled exception while starting server:")
            raise
    #endregion
    
    #region Handlers
    def attach_controllers(self, *controllers: HandlerController, **kwargs):
        if (not controllers):
            controllers = self.handler_controllers
        
        for host in self.hosts:
            host_handlers = list()
            for _handler_controller in controllers:
                host_handlers.extend(_handler_controller.get_handlers_for(host))
            
            self.add_host_handlers(*host_handlers, host=host, **kwargs)
    
    def add_host_handlers(self, *host_handlers: HandlerType, host: str = None, settings: dict = None, protocol = None, listen_port_part = None, **kwargs):
        if (settings is None):
            settings = self.__settings
        if (protocol is None):
            protocol = self.__host_params['protocol']
        if (listen_port_part is None):
            listen_port_part = self.__host_params['listen_port_part']
        if (host is None):
            for _host in self.hosts:
                self.add_host_handlers(*host_handlers, host=_host, settings=settings, protocol=protocol, listen_port_part=listen_port_part, **kwargs)
                return
        
        host_address = settings.get(f'{host}_address')
        host_pattern = None
        
        if (host_address):
            listen_address = None
        else:
            if (hasattr(self, f'get_{host}_host') and callable(getattr(self, f'get_{host}_host'))):
                _host = getattr(self, f'get_{host}_host')()
                if (isinstance(_host, tuple) and len(_host) == 2):
                    listen_address, host_pattern = _host
                else:
                    listen_address = _host
            else:
                listen_address = \
                    settings.get('listen_address') \
                    or settings.get('listen_ip') \
                    or settings.get('ip') \
                    or socket.gethostbyname(socket.gethostname()) \
            
            host_address = f"{protocol}://{listen_address}{listen_port_part}"
        
        if (host_pattern is None):
            if (listen_address is None):
                host_pattern = '.*'
            else:
                host_pattern = re.escape(listen_address)
        
        self.logger.develop(f"Adding host handlers: '{host_pattern}' => {host_handlers}")
        self.add_handlers(re.compile(host_pattern, re.IGNORECASE), list(host_handlers))
    
    def add_router_handlers(self, *handlers: HandlerType, **kwargs):
        self.logger.develop(f"Adding router handlers: => {handlers}")
        for h in handlers:
            self.__router.add_handler(*h, **kwargs)
    
    def add_handlers(self, host_pattern: Union[str, RegExpType] = None, host_handlers: List[Union[HandlerType, Rule]] = None, **kwargs):
        assert host_handlers is not None
        
        if (self.__router is not None):
            self.add_router_handlers(*host_handlers, **kwargs)
        if (host_pattern is not None):
            super().add_handlers(host_pattern, host_handlers)
    
    def find_handler(self, request: HTTPServerRequest, **kwargs) -> Optional[HTTPMessageDelegate]:
        self.normalize_request(request)
        self.logger.develop(f"Finding handler for request '{request}'")
        return super().find_handler(request, **kwargs)
    
    def normalize_request(self, request: HTTPServerRequest) -> HTTPServerRequest:
        path, sep, query = request.uri.partition('?') # type: str, str, str
        _new_path = '/'.join([''] + [_part for _part in path.split('/') if _part])
        if (path.endswith('/')):
            _new_path += '/'
        if (_new_path != path):
            _new_uri = _new_path + sep + query
            self.logger.trace(f"Rewriting uri: '{path}' => '{_new_path}'")
            request.uri = _new_uri
            request.path = _new_path
        return request
    #endregion
    
    #region Logging
    def log_request(self, *args, **kwargs):
        method = self._log_request_new if (self.new_format_access_log) else super().log_request
        return method(*args, **kwargs) 
    
    def _log_request_new(self, handler: Logged_RequestHandler):
        if ('log_function' in self.settings):
            self.settings['log_function'](handler)
            return
        
        _logger = self.__access_logger
        if (handler.get_status() < 400):
            log_method = _logger.info
        elif (handler.get_status() < 500):
            log_method = _logger.warning
        else:
            log_method = _logger.error
        request_time = 1000.0 * handler.request.request_time()
        log_method("%d %s %.2fms", handler.get_status(), handler._request_summary(), request_time, style='%', prefix='resp', request_id=handler.request_id)
    #endregion
